/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* 
 * Wrapper functions to use GNU TLS library to protect multi-stream SCTP association.
 *
 *
 * This wrapper creates N+1 threads, where N is the number of streams in the association.
 *
 * To use this wrapper one must:
 *  - first, create a one-to-one SCTP association with another peer (draft-ietf-tsvwg-sctpsocket-17):
 *     -> Create the socket        :  int socket(PF_INET6, SOCK_STREAM, IPPROTO_SCTP);
 *     -> Set options (streams,...):  setsockopt(SCTP_INITMSG, ...)
 *		 Some options are mandatory for TLS, such as SCTP_DISABLE_FRAGMENTS=OFF
 *		 Some others are useful, such as SCTP_EVENTS.
 *     -> (server) Bind the socket :  int bind( sd, addr, addrlen);
 *     ->                   and/or :  int sctp_bindx( sd, addrs, addrcnt, flags);
 *     -> (server) Listen for cnx  :  int listen( sd,  backlog);
 *     -> (server) Accept clients  :  new_sd = accept( sd, *addr, *addrlen);
 *     -> (client) Connect to serv :  int connect( sd, addr, addrlen);
 *    Note that the helper functions tls_sctp_create_server / tls_sctp_accept / tls_sctp_connect can be used for this.
 * 
 *  - Initialize the wrapper context for this association:
 *     int tls_sctpw_init ( new_sd, streams, &psctx );
 *
 *  - Handshake. One may need to change content to adapt to one's need for security (authentication of certificates, ...)
 *	int tls_sctpw_handshake( tls_sctpw_ctx * sctx );
 *
 *  - Send and receive data on the association, protected with TLS.
 *      int tls_sctpw_send( tls_sctpw_ctx * sctx, int stream, void * data, size_t len );
 *      int tls_sctpw_recv( tls_sctpw_ctx * sctx, int *stream, void ** data, size_t *len );
 *
 *  - Terminate after use, either properly or on error.
 *      int tls_sctpw_close( tls_sctpw_ctx ** sctx );
 *      int tls_sctpw_destroy( tls_sctpw_ctx ** sctx );
 *
 *  - Finally, close or abord the SCTP association
 *      int close(int sd);
 *      see SO_LINGER sockopt to abord...
 */

#ifndef _SCTP_WRAPPER_H
#define _SCTP_WRAPPER_H

#include <sys/socket.h>
#include <sys/types.h>

/****************************************************************************
*         SCTP helper functions, to manage SCTP connections                 *
****************************************************************************/ 
/*
 * FUNCTION:	tls_sctp_create_server
 *
 * PARAMETERS:
 *  sock    : (out) On return, the identifier of the bound socket is stored here.
 *  port    : (in) The port on which the server is listening for connections.
 *  streams : (in) The maximum number of pairs of streams that we will use.
 *
 * DESCRIPTION: 
 *  Creates a listening SCTP bound socket, with some default options.
 *
 * RETURN VALUE:
 *   0 : The socket has been created, bound, and is listening for connections.
 *  !0 : an error occurred (standard errno are returned, such as ENOTSUP)
 */
int tls_sctp_create_server( int * sock, uint16_t port, int streams );
		 
/*
 * FUNCTION:	tls_sctp_accept
 *
 * PARAMETERS:
 *  sock    : The listening server socket from which to accept clients.
 *  new_sd  : on return, the new socket for this client.
 *  streams : the number of streams that can be used for TLS with this client.
 *
 * DESCRIPTION: 
 *  Accepts an incoming connection and get the number of bi-lateral streams.
 *
 * RETURN VALUE:
 *   0 : A new client has arrived.
 *  !0 : an error occurred (standard errno are returned)
 */
int tls_sctp_accept( int sock, int * new_sd, int *streams );

/*
 * FUNCTION:	tls_sctp_connect
 *
 * PARAMETERS:
 *  sock    : The client socket is stored here on return.
 *  addr    : address of the server, that is passed to connect() function.
 *  addrlen : length of addr structure.
 *  streams : (in/out) the number or streams desired / usable for TLS.
 *
 * DESCRIPTION: 
 *  Connect to a server and retrieve the number of usable pairs of streams.
 *
 * RETURN VALUE:
 *   0 : The client is connected.
 *  !0 : an error occurred (standard errno are returned)
 */
int tls_sctp_connect( int *sock, const struct sockaddr *addr, socklen_t addrlen, int *streams );

/*
 * FUNCTION:	tls_sctp_send
 *
 * PARAMETERS:
 *  socket  : The socket on which to send a message.
 *  streamid: The stream on which the message must be sent.
 *  data    : pointer to the data that must be sent.
 *  len     : length of the data buffer.
 *
 * DESCRIPTION: 
 *  Send some data over an SCTP association on a given stream.
 * The return values and error codes are the same as sendmsg().
 *
 * RETURN VALUE:
 *   -1 : error (errno is set)
 *  !-1 : number of bytes sent.
 */
ssize_t tls_sctp_send(int socket, int streamid, const void * data, size_t len);

/* Some events that may be received as notifications in next function */
typedef enum {
	RCV_NO_EVENT = 0,	/* no special event was received */
	RCV_CHANGE,		/* Notified of a change on association or address */
	RCV_ERROR,		/* Notified of an error in sending or remote handling */
	RCV_SHUTDOWN,		/* Notified of a shutdown */
	RCV_MAX			/* the last defined value */
} tls_sctp_recvevents;

/*
 * FUNCTION:	tls_sctp_recv
 *
 * PARAMETERS:
 *  socket  : The socket on which to receive the next message.
 *  event   : if an event is received, it is stored here.
 *  streamid: if relevant, the stream id on which data or event is received.
 *  data    : a buffer containing the received data.
 *  len     : the size of the data buffer
 *  init    : the initial size of the buffer
 *
 * DESCRIPTION: 
 *  Receive data and notifications on a SCTP socket.
 * The data buffer is allocated and must be freed after use.
 * The return values and error codes are the same as recvmsg().
 *
 * RETURN VALUE:
 *  -1 : an error occurred
 *   0 : an event was received or shutdown is in progress.
 *  >0 : length of data written in the data buffer.
 */
ssize_t tls_sctp_recv(int socket, tls_sctp_recvevents * event, int * streamid, void ** data, size_t *len, size_t init);

#endif /* _SCTP_WRAPPER_H */

#ifndef _GNUTLS_SCTP_WRAPPER_H
#define _GNUTLS_SCTP_WRAPPER_H

#include <gnutls/gnutls.h>
#include <gcrypt.h>


/****************************************************************************
*        TLS over SCTP wrapper library -- to use GNU TLS over SCTP          *
****************************************************************************/ 
/* A wrapper context object, to be used in the functions of this file */
typedef void tls_sctpw_ctx;

/*
 * FUNCTION:	tls_sctpw_init
 *
 * PARAMETERS:
 *  sock    : The socket object representing the (one-to-one style) SCTP association.
 *  streams : The number of bi-lateral streams in this association.
 *  sctx    : On success, the wrapper context is stored here.
 *  client  : if 0, this side is server, otherwise this side is client
 *  credtype: type of credentials, GNUTLS_CRD_ANON, GNUTLS_CRD_SRP or GNUTLS_CRD_CERTIFICATE
 *  creds   : the credentials to present to the remote peer
 *  priocb  : optional, a callback to set the priority of algorithms in the session
 *            prototype is:     int priocb (gnutls_session_t * session);
 *
 * DESCRIPTION: 
 *  Creates a wrapper context for one SCTP multi-stream association.
 *
 * RETURN VALUE:
 *   0 : The context has been created sucessfully.
 *  !0 : an error occurred (standard errno are returned, such as ENOMEM)
 */
int tls_sctpw_init( int sock, int streams, tls_sctpw_ctx ** sctx, int client, gnutls_credentials_type_t credtype, void ** creds, int (*priocb)(gnutls_session_t *) );

/*
 * FUNCTION:	tls_sctpw_handshake
 *
 * PARAMETERS:
 *  sctx    : The wrapper context of the connection where handshake must be started.
 *  checkcb : a callback to verify the credentials of the remote peer.
 *  parm    : argument to pass to the checkcb function
 *
 * DESCRIPTION: 
 *  Will handshake on all pairs of streams of the association.
 * A full handshake is done on each pair. This can be improved by resuming sessions
 * on other streams after the first handshake is achieved.
 * The checkcb callback can be provided to verify the credentials of the remote peer.
 * the prototype is:
 *     int checkcb ( gnutls_session_t session, void * parm)
 * The function must return 0 if credentials are valid, or a standard errno value otherwise.
 *
 * RETURN VALUE:
 *   0 : Ok.
 *  !0 : An error occurred (standard errno are returned)
 */
int tls_sctpw_handshake( tls_sctpw_ctx * sctx, int (*checkcb)(gnutls_session_t session, void * parm), void * parm );

/*
 * FUNCTION:	tls_sctpw_send
 *
 * PARAMETERS:
 *  sctx    : The wrapper context of the connection.
 *  stream  : The stream id on which the data must be sent.
 *  data    : The data that must be sent.
 *  len     : the size of data to send.
 *
 * DESCRIPTION: 
 *  Will send data over the association, protected by TLS.
 * If the stream is "-1", the next stream for the previously sent message is used.
 * Otherwise, the stream must be between 0 and the number of streams of the association.
 * Note that despite we are using SCTP, the message boundaries are not preserved with TLS.
 *
 * RETURN VALUE:
 *   0 : Ok.
 *  !0 : An error occurred (standard errno are returned)
 */
int tls_sctpw_send( tls_sctpw_ctx * sctx, int stream, unsigned char * data, size_t len );

/*
 * FUNCTION:	tls_sctpw_recv
 *
 * PARAMETERS:
 *  sctx    : The wrapper context of the connection.
 *  stream  : on return, the stream id on which the data was received.
 *  data    : on return, pointer to received data. This must be freed after use.
 *  len     : on return, the length of the received data.
 *
 * DESCRIPTION: 
 *  Receive data from any stream on an SCTP association.
 * The buffer is dynamically allocated and must be freed after use.
 *
 * RETURN VALUE:
 *   0 : Ok.
 *  !0 : An error occurred (standard errno are returned)
 */
int tls_sctpw_recv( tls_sctpw_ctx * sctx, int *stream, void ** data, size_t *len );

/*
 * FUNCTION:	tls_sctpw_close
 *
 * PARAMETERS:
 *  sctx    : The wrapper context of the connection.
 *
 * DESCRIPTION: 
 *  Terminates the TLS sessions properly, then free the wrapper resources.
 *
 * RETURN VALUE:
 *   0 : Ok.
 *  !0 : An error occurred (standard errno are returned)
 */
int tls_sctpw_close( tls_sctpw_ctx ** sctx );

/*
 * FUNCTION:	tls_sctpw_destroy
 *
 * PARAMETERS:
 *  sctx    : The wrapper context of the connection.
 *
 * DESCRIPTION: 
 *  Terminates the TLS sessions abruptly, and free the wrapper resources.
 *  Use this when the connection is broken and no data can be exchanged anymore with remote peer.
 *
 * RETURN VALUE:
 *   0 : Ok.
 *  !0 : An error occurred (standard errno are returned)
 */
int tls_sctpw_destroy( tls_sctpw_ctx ** sctx );



#endif /* _GNUTLS_SCTP_WRAPPER_H */
