/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Yacc configuration parser.
 *
 * This file defines the grammar for parsing the configuration file of waaad.
 * Note that each extension has a separate independant configuration file.
 *
 * Note : This module is NOT thread-safe. All processing must be done from one thread only.
 */

/* For development only : */

%debug 
%error-verbose

/* The parser receives a configuration pointer as parameter */
%parse-param {waaad_conf_t * pconfig}

/* Keep track of location */
%locations 
%pure-parser

%{
#include "waaad-internal.h"
#include "waaad.tab.h"	/* bison is not smart enough to define the YYLTYPE before including this code, so... */

#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>

/* The Lex parser prototype */
int waaadlex(YYSTYPE *lvalp, YYLTYPE *llocp);

/* This function checks a string value is a valid network locator */
static int is_valid_locator( char * candidate, sSS *ss, waaad_conf_t * pconfig ) 
{
	int ret=0;
	struct addrinfo *ai = NULL, hint;
	
	memset(&hint, 0, sizeof(hint));
	
	hint.ai_family = AF_UNSPEC;
	
	if (pconfig->pub.disable_inet6)
		hint.ai_family = AF_INET;
	if (pconfig->pub.disable_inet4)
		hint.ai_family = AF_INET6;
		
	/* TRACE_DEBUG(FULL, "Resolving '%s'...", candidate); */
	ret = getaddrinfo(candidate, NULL, &hint, &ai);
	if (ai) {
		/*
		TRACE_DEBUG(FULL, "Result:");
		TRACE_DEBUG(FULL, "  ai_flags    : %x", ai->ai_flags);
		TRACE_DEBUG(FULL, "  ai_family   : %d", ai->ai_family);
		TRACE_DEBUG(FULL, "  ai_socktype : %d", ai->ai_socktype);
		TRACE_DEBUG(FULL, "  ai_protocol : %d", ai->ai_protocol);
		TRACE_DEBUG(FULL, "  ai_addrlen  : %d", ai->ai_addrlen);
		TRACE_DEBUG(FULL, "  ai_addr     : .sa_family = %d (%s)", ai->ai_addr->sa_family, ai->ai_addr->sa_family == AF_INET ? "AF_INET" :
													ai->ai_addr->sa_family == AF_INET6 ? "AF_INET6" :
														"???");
		TRACE_DEBUG(FULL, "  ai_canonname: %s", ai->ai_canonname ? ai->ai_canonname:"(null)");
		TRACE_DEBUG(FULL, "  ai_next     : %p", ai->ai_next);
		*/
		if (ss) {
			memset(ss, 0, sizeof(sSS));
			memcpy(ss, ai->ai_addr, ai->ai_addrlen);
		}
		freeaddrinfo(ai);
	}
	if (ret != 0) {
		log_error("Error resolving network locator '%s': %s.\n", candidate, gai_strerror(ret));
		return 0;
	}

	/* Ok this candidate is valid */
	return 1;
}

/* This function checks a string value can be a DiameterId (== a fqdn) */
static int is_valid_fqdn( char * candidate, sSS *ss, waaad_conf_t * pconfig ) 
{
	/* What is a valid FQDN? 
	   Difficult question... maybe use getaddrinfo with flag AI_CANONNAME ?
	   Let's be permissive here and trust administrators to configure properly their Diameter daemon ^^'
	 */
	
	/* We search for a '.' to reject some aliases and local names */
	if (!strchr(candidate, '.')) {
		log_error("The string '%s' is not a valid fully-qualified domain name (fqdn).\n", candidate);
		return 0;
	}
	
	/* Ok now check this is a valid network locator */
	return is_valid_locator(candidate, ss, pconfig);
}

/* This function checks a string value is a valid filename */
static int is_valid_file( char * candidate ) 
{
	int ret;
	struct stat buffer;
	
	ret = stat(candidate, &buffer);
	if (ret != 0) {
		log_error("Error on file '%s': %s.\n", candidate, strerror(errno));
		return 0;
	}

	/* Ok this candidate is valid */
	return 1;
}

/* Global variable to store a sockaddr checked by is_valid_addr or fqdn */
static sSS _ss;

static int have_ip4 = 0; /* have some ip4 been added already? */
static int have_ip6 = 0; /* have some ip6 been added already? */

/* This function converts a string into an IP or IPv6 address */
static int is_valid_addr( char * candidate, waaad_conf_t * pconfig) 
{
	int ret=0;
	
	/* Clean the storage area */
	memset(&_ss, 0, sizeof(_ss));
	
	/* Search for a ":" in the string */
	if (!strchr(candidate, ':')) {
		sSA4 * sin = (sSA4 *) &_ss;
		/* Assume this is IPv4 */
		CHECK_SYS_DO(  ret = inet_pton(AF_INET, candidate, &sin->sin_addr.s_addr), return 0 );
		if (ret == 1) {
			if (pconfig->pub.disable_inet4) {
				log_error("Cannot use the address '%s' since NO_IP4 was defined\n", candidate);
				return 0;
			}
			have_ip4++;
			sin->sin_family = AF_INET;
			sin->sin_port = htons(pconfig->pub.local_port);
			return 1; /* the IP is valid */
		} else {
			log_error("Error on IP address '%s': unrecognized.\n", candidate);
		}
	} else {
		sSA6 * sin6 = (sSA6 *)&_ss;
		/* Assume this is IPv6 */
		CHECK_SYS_DO(  ret = inet_pton(AF_INET6, candidate, &sin6->sin6_addr.s6_addr), return 0 );
		if (ret == 1) {
			if (pconfig->pub.disable_inet6) {
				log_error("Cannot use the address '%s' since NO_IP6 was defined\n", candidate);
				return 0;
			}
			have_ip6++;
			sin6->sin6_family = AF_INET6;
			sin6->sin6_port = htons(pconfig->pub.local_port);
			return 1; /* the IPv6 is valid */
		} else {
			log_error("Error on IPv6 address '%s': unrecognized.\n", candidate);
		}
	}

	/* This candidate is not valid */
	return 0;
}

/* This function extends the list of addresses and stores the new value of _ss inside */
static int queue_ss( sSS ** plist )
{
	sSS * new;
	
	if (*plist == NULL) {
		/* If the array has not been alloc'd yet */
		CHECK_MALLOC(  *plist = malloc( sizeof (sSS) * 2 )  );
		new = *plist;
	} else {
		/* otherwise, we need to resize the array to add the new element */
		int cnt = 0;
		new = *plist;
		while (new->ss_family != 0)
			cnt++, new++;
		CHECK_MALLOC(  *plist = realloc(*plist, sizeof(sSS) * (cnt + 2))  );
		new = (*plist) + cnt;
	}
	
	/* Now, new points to the beginning of the location where the _ss must be stored */
	memcpy(new, &_ss, sizeof(sSS));
	memset(new+1, 0,  sizeof(sSS));
	return 0;
}

/* Function to report the errors */
void yyerror (YYLTYPE *ploc, waaad_conf_t * pconfig, char const *s)
{
	if (ploc->first_line != ploc->last_line)
		log_error("%s:%d.%d-%d.%d : %s\n", pconfig->filepath, ploc->first_line, ploc->first_column, ploc->last_line, ploc->last_column, s);
	else if (ploc->first_column != ploc->last_column)
		log_error("%s:%d.%d-%d : %s\n", pconfig->filepath, ploc->first_line, ploc->first_column, ploc->last_column, s);
	else
		log_error("%s:%d.%d : %s\n", pconfig->filepath, ploc->first_line, ploc->first_column, s);
}

/* Global variable to store temporarily extension information */
static struct _ext_detail {
	char * fname;
	char * cname;
} ext_detail={NULL, NULL};

/* Global variable to store temporarily peer information */
static struct {
	char   * locator;
	sSS	 peerss;
	uint16_t port;
	uint16_t sctp_streams;
	char     flags;
} peer_detail = {.locator = NULL};

/* Flags values; 0 is default behavior */
#define PEER_NO_SEC	0x1
#define PEER_TCP	0x2

%}

/* Values returned by lex for token */
%union {
	char 		*string;	/* The string is allocated by strdup in lex.*/
	uint16_t	 u16;		/* Store small values such as TCP port or timers */
	int		 integer;	/* Store integer values */
}

/* In case of error in the lexical analysis */
%token 		LEX_ERROR

/* A quoted string */
%token <string>	QSTRING
/* Strings subtypes */
%type <string>	FQDN		/* This is a fqdn. We check that the syntax is correct. */
%type <string>	FILENAME	/* This is a filename. existence of the file is checked. */
%type <string>  IPADDRESS	/* This is an IP. the numerical value is stored in _ss. */

/* An integer value */
%token <integer> INTEGER

/* In the "general" section */
%token 		SECTION_GENERAL
%token		GEN_LOG
%type	<u16>	gen_log_type
%token		GEN_LOG_CONSOLE
%token		GEN_LOG_SYSLOG
%token		GEN_DIAMID
%token		GEN_PORT
%type	<u16>	gen_port_val
%token		GEN_TCTIMER
%token		GEN_TWTIMER
%token		GEN_DISABLERELAY
%token		GEN_DISABLETCP
%token		GEN_DISABLEIP4
%token		GEN_DISABLEIP6
%token		GEN_PRIMARY
%token		GEN_SECONDARY
%token		GEN_SCTP_STREAMS
%type	<u16>	gen_sctp_streams_val

/* In the "extension" section */
%token 		SECTION_EXTENSION
%token <string>	EXT_NAME
%token		EXT_FILENAME
%token		EXT_CONFFILE

/* In the "peer" section */
%token 		SECTION_PEER
%token		PEER_URI
%token		PEER_SCHEME_SEC
%token		PEER_SCHEME_NOSEC
%token <string>	PEER_LOCATOR
%token <u16>	PEER_PORT
%token 		PEER_TRANS_SCTP
%token 		PEER_TRANS_TCP

/* -------------------------------------- */
%%

	/* The grammar definition - Sections blocs. */
conffile:		  general_section extensions_section peers_section
			| general_section peers_section extensions_section
			;

	/* These types are used several times: */
FQDN:			QSTRING
			{
				/* Verify this is a valid FQDN */
				if (!is_valid_fqdn($1, &_ss, pconfig)) {
					yyerror (&yylloc, pconfig, "An error was detected on a fqdn, aborting...");
					YYERROR;
				}
				$$ = $1;
			}
			;
				
FILENAME:		QSTRING
			{
				/* Verify this is a valid file */
				if (!is_valid_file($1)) {
					yyerror (&yylloc, pconfig, "An error was detected on a file name, aborting...");
					YYERROR;
				}
				$$ = $1;
			}
			;
				
IPADDRESS:		QSTRING
			{
				/* Verify this is a valid IP */
				if (!is_valid_addr($1, pconfig)) {
					yyerror (&yylloc, pconfig, "An error was detected on an IP address, aborting...");
					YYERROR;
				}
				/* _ss is set with the IP numerical value during the check */
				/* Return the string anyway, to be freed */
				$$ = $1;
			}
			;
				

	/* Define the [general] section grammar */
general_section:	SECTION_GENERAL general_content
			;

general_content:	/* empty is allowed */
			| general_content gen_log 
			| general_content gen_diamid
			| general_content gen_port
			| general_content gen_sctp_streams
			| general_content gen_tctimer
			| general_content gen_twtimer
			| general_content gen_disablerelay
			| general_content gen_disabletcp
			| general_content gen_disableip4
			| general_content gen_disableip6
			| general_content gen_primary
			| general_content gen_secondary
			;

gen_log:		GEN_LOG '=' gen_log_type ';'
			{
				/* Save the new log type in the configuration. We'll set it only after conf file is parsed */
				pconfig->log_to = (int) $3;
			}
			;

gen_log_type:		/* Terminals rules */
			  GEN_LOG_SYSLOG
			{
				/* Save the syslog flag */
				$$ = LOG_TO_SYSLOG;
			}
			| GEN_LOG_CONSOLE
			{
				/* Save the console flag */
				$$ = LOG_TO_CONSOLE;
			}
			/* Recursive rules */
			| gen_log_type ',' GEN_LOG_SYSLOG
			{
				/* Add the syslog flag */
				$$ = $1 | LOG_TO_SYSLOG;
			}
			| gen_log_type ',' GEN_LOG_CONSOLE
			{
				/* Add the console flag */
				$$ = $1 | LOG_TO_CONSOLE;
			}
			;

gen_diamid:		GEN_DIAMID '=' FQDN ';'
			{
				/* Save this string as the peer diameter id */
				if (pconfig->pub.diameter_identity) {
					free(pconfig->pub.diameter_identity);
					pconfig->pub.diameter_identity = NULL;
				}
				pconfig->pub.diameter_identity = $3;
			}
			| GEN_DIAMID '=' error ';'
			{
				log_error("A Diameter-Id MUST be a fqdn. Exiting...\n");
				YYERROR;
			}
			;

gen_port:		GEN_PORT '=' gen_port_val ';'
			{
				pconfig->pub.local_port = $3;
			}
			;
			
gen_port_val:		INTEGER
			{
				if (( $1 <= 0) || ($1 > 65535)) {
					log_error("The LocalPort value (%d) is invalid. Exiting...\n", $1);
					YYERROR;
				}
				$$ = (uint16_t) $1 ;
			}
			;

gen_sctp_streams:	GEN_SCTP_STREAMS '=' gen_sctp_streams_val ';'
			{
				pconfig->pub.sctp_streams = $3;
			}
			;
			
gen_sctp_streams_val:	INTEGER
			{
				if (( $1 <= 0) || ($1 > 65535)) {
					log_error("The sctp_streams value (%d) is invalid. Exiting...\n", $1);
					YYERROR;
				}
				$$ = (uint16_t) $1 ;
			}
			;

gen_tctimer:		GEN_TCTIMER '=' INTEGER ';'
			{
				pconfig->pub.tctimer = $3;
			}
			;

gen_twtimer:		GEN_TWTIMER '=' INTEGER ';'
			{
				if ($3 >= 6) {
					pconfig->pub.twtimer = $3;
				} else {
					log_error("The TwTimer value (%d) is invalid. It must be >= 6. Exiting...\n", $3);
					YYERROR;
				}
			}
			;

gen_disablerelay:	GEN_DISABLERELAY '=' INTEGER ';'
			{
				pconfig->pub.disable_relay = $3;
			}
			;

gen_disabletcp:		GEN_DISABLETCP ';'
			{
				pconfig->pub.disable_tcp = 1;
			}
			;

gen_disableip4:		GEN_DISABLEIP4 ';'
			{
				if (pconfig->pub.disable_inet6) {
					yyerror (&yylloc, pconfig, "NO_IP6 and NO_IP4 can not be used together.");
					YYERROR;
				}
				if (have_ip4) {
					yyerror (&yylloc, pconfig, "A primary or secondary IP address was defined, but NO_IP4 appears.");
					YYERROR;
				}
				pconfig->pub.disable_inet4 = 1;
			}
			;

gen_disableip6:		GEN_DISABLEIP6 ';'
			{
				if (pconfig->pub.disable_inet4) {
					yyerror (&yylloc, pconfig, "NO_IP4 and NO_IP6 can not be used together.");
					YYERROR;
				}
				if (have_ip6) {
					yyerror (&yylloc, pconfig, "A primary or secondary IPv6 address was defined, but NO_IP6 appears.");
					YYERROR;
				}
				pconfig->pub.disable_inet6 = 1;
			}
			;
			
gen_primary:		GEN_PRIMARY '=' IPADDRESS ';'
			{
				/* queue the sa in _ss into pconfig->pub.local_addr_pri_sa */
				if (queue_ss(&pconfig->pub.local_addr_pri_sa) != 0) {
					yyerror (&yylloc, pconfig, "An error occurred while saving an address, aborting...");
					YYERROR;
				}
				
				
				/* we don't need the string anymore */
				free($3);
			}
			;

gen_secondary:		GEN_SECONDARY '=' IPADDRESS ';'
			{
				/* queue the sa in _ss into pconfig->pub.local_addr_sec_sa */
				if (queue_ss(&pconfig->pub.local_addr_sec_sa) != 0) {
					yyerror (&yylloc, pconfig, "An error occurred while saving an address, aborting...");
					YYERROR;
				}
				
				/* we don't need the string anymore */
				free($3);
			}
			;


	/* Define the [extensions] section grammar */
extensions_section:	SECTION_EXTENSION extensions_content
			;

extensions_content:	/* empty is allowed */
			| extensions_content extension_detail
			;

extension_detail:	{
				/* Reset the ext_detail global before processing the rule */
				if (ext_detail.fname) { free(ext_detail.fname); ext_detail.fname = NULL; }
				if (ext_detail.cname) { free(ext_detail.cname); ext_detail.cname = NULL; }
			}
			QSTRING '=' '{' extension_params '}' ';'
			{
				int ret = 0;
				
				/* Add this extension */
				ret = ext_register( ext_detail.fname, ext_detail.cname );
				
				/* Free the strings */
				if (ext_detail.fname) { free(ext_detail.fname); ext_detail.fname = NULL; }
				if (ext_detail.cname) { free(ext_detail.cname); ext_detail.cname = NULL; }
				
				/* Check error */
				if (ret != 0) {
					log_error("Failed to register extension %s: %s\n", $2, strerror(ret));
					YYERROR;
				}
			}
			;

extension_params:	extension_confname extension_filename extension_confname	/* confname appears twice on purpose, to allow any order */
			;

extension_filename:	EXT_FILENAME '=' FILENAME ';'
			{
				if (ext_detail.fname) { free(ext_detail.fname); ext_detail.fname = NULL; }
				ext_detail.fname = $3;
			}
			| EXT_FILENAME '=' error ';'
			{
				log_error("You must provide a valid filename for extension.\n");
				YYERROR;
			}
			;

extension_confname:	/* This one is optional */
			| EXT_CONFFILE '=' FILENAME ';'
			{
				if (ext_detail.cname) { free(ext_detail.cname); ext_detail.cname = NULL; }
				ext_detail.cname = $3;
			}
			| EXT_CONFFILE '=' error ';'
			{
				log_error("If you provide a configuration file for the extension, it must be a valid file.\n"
					  " Note that you may remove the 'conffile' directive if your extension has no configuration.\n");
				YYERROR;
			}
			;

	/* Define the [peers] section grammar */
peers_section:		SECTION_PEER peers_content
			;

peers_content:		/* empty is allowed */
			| peers_content peer_detail
			;

peer_detail:		{
				/* Initialize the peer details */
				if (peer_detail.locator ) { free(peer_detail.locator); }
				memset(&peer_detail, 0, sizeof(peer_detail));
				peer_detail.port = 3868;
				peer_detail.sctp_streams = pconfig->pub.sctp_streams;
			}
			FQDN '=' '{' peer_content '}' ';'
			{
				int ret = 0;
				peer_add_t data;
				
				memset(&data, 0, sizeof(peer_add_t));
				
				/* Copy information from peer_detail */
				
				/* The "no sec" flag */
				if (peer_detail.flags & PEER_NO_SEC)
					data.pa_flags = PA_NOSEC;
				
				/* The peers declared in the configuration file do not expire */
				
				/* Now fill the connection information */
				if (peer_detail.flags & PEER_TCP) {
					data.pa_proto = IPPROTO_TCP;
				} else {
					data.pa_proto = IPPROTO_SCTP;
				}
				
				/* the SA information was resolved in peerss during the "peer_locator" rule, or fqdn */
				if (peer_detail.locator)
					memcpy(&data.pa_ss, &peer_detail.peerss, sizeof(sSS));
				else
					memcpy(&data.pa_ss, &_ss, sizeof(sSS));
					
				switch (data.pa_ss.ss_family) {
					case AF_INET:
						{
							sSA4 * sin = (sSA4 *)&data.pa_ss;
							sin->sin_port = htons(peer_detail.port);
						}
						break;
					case AF_INET6:
						{
							sSA6 * sin6 = (sSA6 *)&data.pa_ss;
							sin6->sin6_port = htons(peer_detail.port);
						}
						break;
					default:
						TRACE_DEBUG(INFO, "Unknown AF: %d", data.pa_ss.ss_family);
						yyerror(&yylloc, pconfig, "Resolved peer to unknown address family...");
						YYERROR;
				}
				
				/* Save the number of streams */
				data.pa_streams = peer_detail.sctp_streams;
				
				/* Now add the peer with its information */
				ret = peer_add($2, strlen($2), &data);
				if (ret != 0) {
					yyerror(&yylloc, pconfig, "An error occurred while adding a peer, aborting...");
					YYERROR;
				}
				
				free($2);
				if (peer_detail.locator ) { free(peer_detail.locator); peer_detail.locator=NULL; }
			}
			;
			
peer_content:		/* empty is allowed */
			| peer_content peer_uri
			| peer_content peer_sctp_streams
			;

peer_uri:		PEER_URI '=' '"' peer_scheme peer_locator peer_port peer_transport '"' ';'
			;

peer_scheme:		PEER_SCHEME_SEC 
			| PEER_SCHEME_NOSEC
			{
				peer_detail.flags |= PEER_NO_SEC;
			}
			;

peer_locator:		PEER_LOCATOR
			{
				/* Check the locator is valid */
				if (!is_valid_locator($1, &peer_detail.peerss, pconfig)) {
					yyerror (&yylloc, pconfig, "An error was detected on a locator, aborting...");
					YYERROR;
				}
				if (peer_detail.locator) 
					free(peer_detail.locator);
				peer_detail.locator = $1;
			}
			;
			
peer_port:		/* empty */
			| PEER_PORT
			{
				peer_detail.port = $1;
			}
			;
			
peer_transport:		/* empty */
			| PEER_TRANS_SCTP
			| PEER_TRANS_TCP
			{
				if (pconfig->pub.disable_tcp) {
					yyerror(&yylloc, pconfig, "Transport is set to TCP but this protocol is disabled in general section.");
					YYERROR;
				}
				peer_detail.flags |= PEER_TCP;
			}
			;

peer_sctp_streams:	GEN_SCTP_STREAMS '=' gen_sctp_streams_val ';'
			{
				peer_detail.sctp_streams = $3;
			}
			;
